import defs.Tree

#This class represents a plan library
class PL(object):
    '''
    classdocs
    '''


    def __init__(self, sigma, NT, R, P):
        '''
        Constructor
        '''
        self._Sigma = sigma     # list of Sigmas
        self._NT = NT           # list of NTs
        self._R = R             # subset of NT
        self._P = P             # list of rules
        
    
    def __repr__(self):
        res = "Plan Library: \nSigma=" + self._Sigma.__repr__() + "\nNT=" + self._NT.__repr__() + "\nR=" + self._R.__repr__() + "\nRules="
        for rule in self._P:
            res += "\t" + rule.__repr__() + "\n"
        return res
    
    def generatingSet(self):
        #res is a set of all leftmost trees deriving from this PL
        res=[]
        
        for goal in self._NT:
            for rule in self._P:
                if goal == rule._A:
                    for tree in self.createTrees(rule):
                        res.append(tree)
        return res                    
        
    def createTrees(self, rule):
        #if rule has only one child 
            #(assuming rule's alpha is Sigma, so there can be no rule like A->B)
            #This can be enforced in the grammar construction
        if 1 == len(rule._alpha):
            child = defs.Tree.Tree("Basic", rule._alpha[0], [], [], self)
            root = defs.Tree.Tree("Complex", rule._A, rule, [child], self)
            return [root]
        
        trees=[]
        i=0     #needed to keep the correct order of nodes (which is a must for the constraints to be correct)
        for child in rule.leftMostChilds():
            #collect all possible derivations of this child to childTrees
            for childRule in self._P:
                if child == childRule._A:
                    #to create the tree, other children should be basic tree nodes
                    otherChildren=self.otherChildrenTrees(rule, child)
                    i=rule.getChildIndex(child)
                    #add all current leftmostTrees to the list, under trees:
                    for tree in self.createTrees(childRule):
                        allChildren = otherChildren[:]
                        allChildren.insert(i,tree)
                        root = defs.Tree.Tree("Complex", rule._A, rule, allChildren, self)
                        trees.append(root)                    
        return trees
    
    def otherChildrenTrees(self, rule, specialChild):
        res=[]
        for child in rule._alpha:
            if child != specialChild:
                tree = defs.Tree.Tree("Basic", child, [], [], self)
                res.append(tree)       
        #print "        otherChildrenTree=", res
        return res   
    
    def createPFFG(self):
        PFFG=[]
        gl=self.generatingSet()
        for tree in gl:
            s = tree.treeToList()
            foot = tree.isLeftMostTree()
            footIndex = s.index(foot)
            s.remove(foot)
             
            #make constraints list according to list without the foot
            
            # simply don't add this constraint to the new list of constraints
            if len(s) < 2:
                order = []
            #else
            order = []
            tempOrder = tree.treeConstraints()
            for cons in tempOrder:
                if cons[0] == footIndex or cons[1] == footIndex:
                    continue
                cons0 = (cons[0]-1 if cons[0] > footIndex else cons[0])
                cons1 = (cons[1]-1 if cons[1] > footIndex else cons[1])
                order.append((cons0, cons1))
                                
            #create the new rule for the PFFG
            newRule = defs.Rule.Y_Rule(foot, tree._root, s, order)
            PFFG.append(newRule)
            
        return PFFG