import defs.Sigma

class Rule(object):
    '''
    classdocs
    '''


    def __init__(self, A, alpha=[], order=[]):
        '''
        Constructor
        '''
        self._A=A           #this item is either of type Sigma or NT
        self._alpha=alpha   #this is a list of NTs and Sigmas
        self._order=order   #each item in this list is (i,j) which means the ith child must come before the jth chikd
                                #both i and j are of type int
                                
    def __repr__(self):
        res = self._A.get() + " -> "
        for child in self._alpha:
            res += child.get() + " "       
        res += "| [ "
        for cons in self._order:
            res += self.constraintWithLetters(cons)
        res += "]"      
        return res
    
    def constraintWithLetters(self, cons):
        left = self._alpha[cons[0]].get()
        right = self._alpha[cons[1]].get()
        return "(" + left + "," + right + ") "
        
    # Return all leftmost childs of the rule    
    def leftMostChilds(self):
        res = []
        i=0
        
        if self._alpha == None:
            return []
        
        for ch in self._alpha:            
            if not self.hasConstraint(i):
                res.append(ch)
            i+=1
                
        return res;
    
    # Return tuples of leftmost childs with their index
    def leftMostChildsWithNums(self):
        res = []
        i=0
        
        if self._alpha == None:
            return []

        for ch in self._alpha:            
            if not self.hasConstraint(i):
                res.append((ch,i))
            i+=1
                
        return res;
    
    # Return all right childs of the rule
    def rightChilds(self):
        right = self._alpha[:]
        for ch in self.leftMostChilds():
            right.remove(ch)
        return right
    
    # service function for leftmostChilds
    def hasConstraint(self, index):
        for cons in self._order:
            if cons[1] == index:
                return True;
        return False;
    
    
    #return the first index of child in alpha (children's list), or throws an error if there is no such child
    def getChildIndex(self, child):
        return self._alpha.index(child)
    
    # Gets a node from a tree and return true iff the node can represent a leftmost implementation of this rule
    def matchLeftTree(self, node):
        if node == ():
            return False
        
        #cond 3.b
        if node._rule != [] and node._rule != self:
            return False
        
        #cond 3.c
        for rChild in self.rightChilds():
            if () == node.getChild(rChild):
                return False
            elif not node.getChild(rChild).isLeaf():
                return False
        
        #cond 3.d
        foot=();
        for lChild in self.leftMostChilds():
            #print "    lchild=", lChild

            # See if lChild present in the tree
            possibleFooter = node.getChild(lChild);
            
            # if this happens, it means the tree cannot be represented by this rule, since lChild does not appear in this tree
            if () == possibleFooter:
                #print " lChild or rule does not appear in the tree"
                return False
            
            # if this branch have children of its own
            if not possibleFooter.isLeaf():
                    
                #check that at most one lChild has children of its own
                if () != foot:
                    return False
                
                #Some other lChild already have children        
                else:
                        foot=possibleFooter.isLeftMostTree()
        
        # Special case, where tree is leftMost, but none of the Non-terminal leaves have children of its own that can be foot
        if () == foot:
            return True
        
        #If passed all conds up to here          
        return foot  
    
  
#
#    This class represents rules in "YAPPR" style - where each rule is of the form: pid: <a,B>-> string : constraints
#    (Buit accoding to Definition 5 in article)
#      
class Y_Rule(object):
    
    ruleAmount=0
       
    def __init__(self, foot, root, alpha=[], order=[]):
        '''
        Constructor
        '''
        self._foot=foot           #this item is of type Sigma
        self._root=root           #this item is of type NT
        self._alpha=alpha   #this is a list of NTs and Sigmas
        self._order=order   #each item in this list is (a,b) which means a must come before b
                                #both a and b are of type NT or Sigma
        Y_Rule.ruleAmount+=1
        self._pid=Y_Rule.ruleAmount
                                
    def __repr__(self):
        res = str(self._pid) + ": <" + str(self._foot.get()) + "," + str(self._root.get()) +"> -> " + str(self._alpha) + " : " + str(self._order)
        return res   