import defs.NT
import defs.Sigma

class Tree(object):
    '''
    classdocs
    '''


    def __init__(self, actionType, root, rule=[], children=[], PL=[]):
        '''
        Constructor
        '''
        if actionType == "Complex":
            self._rule = rule                         #type Rule
            self._root = root                       #either NT or Sigma
            self._children = children               #list of nodes
            self._PL = PL                           #Plan Library
        
        elif actionType == "Basic":
            self._rule = ()                 #always null for node representing an atomic action
            self._root = root               #type NT or Sigma
            self._children = []             #always null for node representing an atomic action
            self._PL = PL                   #Plan Library
        
        else:
            print "Unknown Node type"
    
    # Try to print the tree nicely
    def __repr__(self, depth=""):
        if [] == self._children:
            return depth+self._root.get()+"\n"
        res = depth+self._root.get()+"\n"
        for child in self._children:
            res+=child.__repr__(depth+"  ")
        return res
         
                
    # Return true if this node has no children    
    def isLeaf(self):
        if self._children == []:
            return True;
        return False
    
    # gets a value (either NT or Sigma) and return the first child with this value as root
    def getChild(self, value):
        for child in self._children:
            if value == child._root:
                return child
        return ()
    
    # Check condition 1+2
    # Condition 1: "every node in T is labeled with a symbol from Sigma union NT"
    # Condition 2: "every interior node in T is labeled with a symbol from NT"
    def isEveryNodeCorrect(self):
        if defs.NT.NT == type(self._root) or defs.Sigma.Sigma == type(self._root):
            if self._children:
                return True
            elif defs.Sigma.Sigma == type(self._root):
                return False
            else:
                for subTree in self._children:
                    if not subTree.isEveryNodeCorrect():
                        return False
                    else:
                        return True
        else:
            return False

    # Service function for allInteriorCorrect
    # checks if node matches interior node rules for its self._rule
    # if so, returns its foot, else return false
    def matchRule(self):
        if () == self._rule:
            return self._root
        if [] == self._rule:
            for rule in self._PL._P:
                isMatchRule = rule.matchLeftTree(self)
                if isMatchRule:
                    self._rule = rule
                    return isMatchRule
            #print "Error: Tree's rule does not match any rule in PL"
            return False
            
        # else - matched to rule before    
        return self._rule.matchLeftTree(self)


    # Check condition 3: "if an interior node, n(=self), labeled A(=root) in T has children
    # with labels beta1...betak(=self.children.roots()), then..."
    def allInteriorCorrect(self):
        if self.isLeaf():
            return self._root
        
        isMatchRule = self.matchRule()
        # if this node matches cond3 (and it's an interior node) - find the foot of this branch
        if isMatchRule:
            foot=False
            for child in self._children:
                childMatchRule = child.allInteriorCorrect()
                #if even child does not match some rule
                if not childMatchRule:
                    return False
                #only one child can be "the foot bearer":              
                if type(childMatchRule) == defs.Sigma.Sigma:
                    #if other child already took this position as the "foot bearer", return False 
                    if foot:
                        return False
                    #else (foot has not been claimed and child ends with Sigma)
                    foot = childMatchRule
                 
            #if exactly one child ends with a foot - return the foot
            #notice that if no child is a foot bearer - returns False
            return foot
        
        #if does not match any rule
        return False
                    
                
                
    
    # Return Foor of tree iff this tree is a leftmost derivation of self._rule
    # Else, return False 
    def isLeftMostTree(self):
        #cond 1+2 - every node is either NT or Sigma, every inner node in NT 
        if not self.isEveryNodeCorrect():
            #print self.__repr__() + "The Tree above does not fill conditions 1+2"
            return False;
        
        foot = self.allInteriorCorrect()
        #cond 3 - each interior node matches all requirements
        if foot == False:
            #print self.__repr__() + "The Tree above does not fill condition 3"
            return False;
        
        return foot
    
    def treeLength(self, withSelf=False):
        if [] == self._children:
            if withSelf:
                return 1
            else:
                return 0           
        res = 0
        for child in self._children:
            res+=child.treeLength(False)
        return res
    
    def distance(self, leftNum, rightNum):
        if leftNum-rightNum == 0:
            return 0
            
        if leftNum-rightNum == 1:
            return self.treeLength()
        
        res = 0
        for child in self._children[leftNum:rightNum-1]:
            res+=child.treeLength()
        return res
                    
    def treeToList(self, withRoot=False):
        if [] == self._children:
            if withRoot:
                return [self._root]
            else:
                return []
            
        res = []
        for child in self._children:
            res.extend(child.treeToList(True))
        return res
   
    # return all constraints in this tree, by 
    def treeConstraints(self, number=0):
        res=[]
        if self._rule ==() or self._rule == []:
            res = []
        else:
            for con in self._rule._order:
                left = self._children[con[0]]
                right = self._children[con[1]]
                if (con[0] < con[1]):
                    leftNum = con[0]+number
                    rightNum = con[1]+self.distance(con[0], con[1])+number
                else:
                    leftNum = con[0]+number
                    rightNum = con[1]+number
                res.extend(self.bubbleConsDownwards(left, right, leftNum, rightNum))
        
        if len(self._children) == 0:
            return res    

        currentIndex=number
        for child in self._children:
            res.extend(child.treeConstraints(currentIndex))
            currentIndex+=child.treeLength(True)
        return res
    
    def bubbleConsDownwards(self, left, right, leftNum=0, rightNum=0):
        res=[]
        if left._children == []:
            if right._children == []:
                res.append((leftNum, rightNum))
            else:
                for additionToIndex in range(len(right._children)):
                    newTup = (leftNum, rightNum+additionToIndex)
                    res.append(newTup)
                        
        else:
            for additionToIndex in range(len(left._children)):
                    res.extend(self.bubbleConsDownwards(left._children[additionToIndex], right, leftNum+additionToIndex, rightNum))
        return res